package net.sf.corn.httpclient;

import java.net.Authenticator;
import java.net.PasswordAuthentication;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * This class is a thread safe implementation of {@link java.net.Authenticator} class. It keeps a thread safe url,credentials pairs map 
 * to use for http client requests.It's permitted to use '*' wildcard in url targets, so requests are authenticated if their targets matches. 
 *  It's also supported to use default credentials to use with any requested url. When an url connection is made by client, 
 *  this class uses credentials from most specific url matches to the most generic one   
 * </br>Usage:</br>
 * <code>
 *   Authenticator.setDefault(new TSAuthenticator());</br>
 *   TSAuthenticator.setCredential("http://www.srv.com/sr/JSONRR","userName","password");</br>
 *   TSAuthenticator.setCredential("http://www.srv.com/*","userName","password");</br>
 *   TSAuthenticator.setDefaultProxyCredential("userName","password");
 * </code>
 * 
 * @author Serhat Dirik
 * @see java.net.Authenticator
 */
public class TSAuthenticator extends Authenticator{

	private static final char PARAMETER_SPECIFIER = '?';
	private static final char PATTERN_SPECIFIER = '*';
	/*
	 * Username password credential
	 */
	private static class UPCredential{
		private transient String uname = null;
		private transient String password = null;
		public UPCredential( String uname, String password) {
			super();
			this.uname = uname;
			this.password = password;
		}
		public String getUname() {
			return uname;
		}
		public String getPassword() {
			return password;
		}
	}
	
	/**
	 * URL Mapped security credentials 
	 *  @author <a href="mailto:serhat.dirik@google.com">Serhat Dirik</a>
	 *
	 */
	private static class TLCredentialsMap extends ThreadLocal<Map<String,UPCredential>>{
		@Override
		protected Map<String, UPCredential> initialValue() {
			return null;
		}
		@Override
		public Map<String, UPCredential> get() {
			return super.get();
		}
		@Override
		public void set(Map<String, UPCredential> value) {
			super.set(value);
		}
	}
	
	private static class TLCredential extends ThreadLocal<UPCredential>{
		@Override
		protected UPCredential initialValue() {
			return null;
		}
		@Override
		public UPCredential get() {
			return super.get();
		}
		@Override
		public void set(UPCredential value) {
			super.set(value);
		}
	}
	
 
	
	private static TLCredentialsMap tlUrlCredentialsMap = new TLCredentialsMap();
	private static TLCredentialsMap tlUrlProxyCredentialsMap = new TLCredentialsMap();
	
	private static TLCredentialsMap tlPatternCredentialsMap = new TLCredentialsMap();
	private static TLCredentialsMap tlPatternProxyCredentialsMap = new TLCredentialsMap();
	
	private static TLCredential tlDefaultCredential = new TLCredential();
	private static TLCredential tlDefaultProxyCredential = new TLCredential();
	
	
	private TSAuthenticator() {
	}
	/**
	 * Sets default security credentials to pass proxy server 
	 * @param uname
	 * @param password
	 */
	public static void setDefaultProxyCredential(String uname,String password){
		tlDefaultProxyCredential.set(new UPCredential(uname,password));
	}
	
	/**
	 * Sets default security credentials 
	 * @param uname
	 * @param password
	 */
	public static void setDefaultCredential(String uname,String password){
		tlDefaultCredential.set(new UPCredential(uname,password));
	}
	
	/**
	 * Sets security credentials to pass proxy server for the given url target
	 * @param uname
	 * @param password
	 */
	public static void setProxyCredentials(String url,String uname,String password){
		if(url == null){
			throw new IllegalArgumentException("Null url value in");
		}
		if(url.indexOf(PATTERN_SPECIFIER) >= 0){
			if(url.indexOf(PATTERN_SPECIFIER) != url.length()){
				throw new IllegalArgumentException("* could be the last character in url value!");
			}
			putInMap(tlPatternProxyCredentialsMap, url, uname, password);
		}else{
			String _url = null;
			if(url.indexOf(PARAMETER_SPECIFIER) > 0){
				_url =  url.toString().substring(0,url.toString().indexOf(PARAMETER_SPECIFIER) -1);
			}else{
				_url = url;
			}
			putInMap(tlUrlProxyCredentialsMap, _url, uname, password);
		}
		
		
		
	}
	
	/**
	 * Sets security credentials for the given url target, the '*' wildcard can be used in url
	 * @param uname
	 * @param password
	 */
	public static void setCredentials(String url,String uname,String password){
		if(url == null){
			throw new IllegalArgumentException("Null url value in");
		}
		if(url.indexOf('*') >= 0){
			if(url.indexOf('*') != url.length()){
				throw new IllegalArgumentException("* could be the last character in url value!");
			}
			putInMap(tlPatternCredentialsMap, url, uname, password);
		}else{
			String _url = null;
			if(url.indexOf(PARAMETER_SPECIFIER) > 0){
				_url =  url.toString().substring(0,url.toString().indexOf(PARAMETER_SPECIFIER) -1);
			}else{
				_url = url;
			}
			putInMap(tlUrlCredentialsMap, _url, uname, password);
		}
	}
	
	/**
	 * Sets this class as the default authenticator in the current JVM
	 */
	@SuppressWarnings({"restriction","unused"})
	public static void  setAsDefault(){
		  Authenticator.setDefault(new TSAuthenticator());
		  try {
		    	//this block is written for suppressing a bug in sun implementation
		    	//In Sun Impl client doesn't authenticate user for each connection , uses cached credentials instead
				Class<?> cacheValClass = Class.forName("sun.net.www.protocol.http.AuthCacheValue");
				sun.net.www.protocol.http.AuthCacheValue.setAuthCache(new sun.net.www.protocol.http.AuthCacheImpl());
			} catch ( Exception e) {
			}
	}
	
	private static void putInMap(TLCredentialsMap map,String url,String uname,String password){
		if(uname == null){
			if(map.get() != null){
				map.get().remove(url);
				if(map.get().size() == 0) map.set(null);
			}
			return;
		}
		UPCredential credential = new UPCredential(uname,password);
		if(map.get() == null){
			map.set(new HashMap<String, TSAuthenticator.UPCredential>());
		}
		map.get().put(url, credential);
	}
 
	private UPCredential findProperCredential(String url, TLCredentialsMap urlMap, TLCredentialsMap patternMap, TLCredential defaultCredential){
		UPCredential credential  =  null;
		credential  = urlMap.get() != null ? urlMap.get().get(url): null;
		
		if(credential == null ){
			Map<String, UPCredential> map = patternMap.get();
			Set<String> keyset = map != null ? map.keySet() : null;
			if(keyset != null){
				for(String pattern: keyset){
					String prefix = pattern.substring(0,pattern.indexOf('*')-1);
					if(url.startsWith(prefix)){
						credential = map.get(pattern);
						break;
					}
				}
			}
		}
		if(credential == null){
			credential = defaultCredential.get();
		}
		return credential;
	}
	
	protected PasswordAuthentication getPasswordAuthentication() {
		String _url = getRequestingURL().toExternalForm();
		if(_url.indexOf(PARAMETER_SPECIFIER) > 0){
			_url =  _url.substring(0,_url.indexOf(PARAMETER_SPECIFIER));
		}else{
			_url = _url.trim();
		}
		
		UPCredential credential  =  null;
		TLCredentialsMap urlMap = null;
		TLCredentialsMap patternMap = null;
		TLCredential defaultCredential = null;
		
		if(getRequestorType().equals(RequestorType.PROXY)){
			urlMap = tlUrlProxyCredentialsMap;
			patternMap = tlPatternProxyCredentialsMap;
			defaultCredential= tlDefaultProxyCredential;
		}else{
			urlMap = tlUrlCredentialsMap;
			patternMap = tlPatternCredentialsMap;
			defaultCredential= tlDefaultCredential;
		}
		credential = findProperCredential(_url, urlMap, patternMap, defaultCredential);
		
		if(credential == null && _url.endsWith("/")){
			credential = findProperCredential(_url.substring(0,_url.length()-1), urlMap, patternMap, defaultCredential);
		}
		
		if(credential != null){
			return new PasswordAuthentication (credential.getUname(),credential.getPassword().toCharArray());
		}else{
			return null;
		}
    }
	
	/**
	 * Resets all credentials that previously set
	 */
	public static void reset(){
		tlDefaultCredential.set(null);
		tlDefaultProxyCredential.set(null);
		tlPatternCredentialsMap.set(null);
		tlPatternProxyCredentialsMap.set(null);
		tlUrlCredentialsMap.set(null);
		tlUrlProxyCredentialsMap.set(null);
	}
}
